/* Copyright (c) 2000  Kevin Sullivan <nite@gis.net>
 *
 * Please refer to the COPYRIGHT file for more information.
 */

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <ncurses.h>
#include <sys/time.h>


#include "defines.h"
#include "cmds.h"
#include "vfuncs.h"
#include "alias.h"


extern vars_t vars[];
extern WINDOW *wchan; 

alias_t *alhead;
sets_t *setl;


char *ins(char *m, char *n, int bp, int ep)
{
  char *st;
  
  st = (char *)malloc(bp+strlen(n)+strlen(m+ep)+1);
  memset(st, 0, bp+strlen(n)+strlen(m+ep)+1);
  
  strncpy(st, m, bp);
  sprintf(st, "%s%s%s", st, n, m+ep);
  
  return(st);
}

void prt(char **tok)
{
  int i;
  char buf[1024];
  
  memset(buf, 0, sizeof(buf));

  for (i=0;tok[i];i++)
  {
    strcat(buf, tok[i]);
    strcat(buf, " ");
  }
  
  buf[strlen(buf)-1] = 0;
  
  wp(wchan, "|%s|\n", buf);
  drw(wchan);
}

unsigned char *dovars(unsigned char *nm)
{
  unsigned char *r=NULL, tp, *t;
  char **vtok=NULL, b[512], **tok;
  int i, j, vcnt=0, cnt, x, y;
  
  if (findset(nm))
    r = strdup(getval(nm, &tp));
  
  if (strchr(nm, '('))
  {
    memset(b, 0, sizeof(b));
    strcpy(b, strchr(nm, '(')+1);
    b[strlen(b)-1] = 0;
    vtok = form_vtoks(b, &vcnt);
  }
  
  if (!vtok)
  {
    for (i=0;i<vcnt;i++)
    {
      if (*vtok[i] == '$')
      {
        t = dovars(vtok[i]+1);
        if (t)
        {
          free(vtok[i]);
          vtok[i] = t;
        }
      }
    }
  }
  else
  {
    for (i=0;i<vcnt;i++)
    {
      tok = form_toks(vtok[i], &cnt);
      tok = fxv(tok, cnt, &cnt);
      for (j=0;j<cnt;j++)
      {
        if (*tok[j] == '$' && !isdigit(tok[j][1]) && tok[j][1] != '+')
        {
          t = dovars(tok[j]+1);
          if (t)
          {
            free(tok[j]);
            tok[j] = t;
          }
        }
      }
      for (j=0;tok[j];j++)
      {
        if (*tok[j] == '$' && tok[j][1] == '+' && !tok[j][2])
        {
          if (j > 0 && tok[j+1] && tok[j-1])
          {
            for (x=1;tok[j-x]&&!(*tok[j-x]);x++);
            for (y=1;tok[j+y]&&!(*tok[j+y]);y++);
            if (!tok[j-x])
              x--;
            if (!tok[j+y])
              y--;
            tok[j] = (char *)realloc(tok[j], strlen(tok[j-x])+strlen(tok[j+y])+1);
            if (strcmp(tok[j-x], "$+"))
              strcpy(tok[j], tok[j-x]);
            else
              *tok[j] = 0;
            if (strcmp(tok[j+y], "$+"))
              strcat(tok[j], tok[j+y]);
            *tok[j-x] = 0;
            *tok[j+y] = 0;
          }
          else
            *tok[j] = 0;
        }
      }
      
      vtok[i] = (char *)realloc(vtok[i], 2048);
      memset(vtok[i], 0, 2048);
      for (j=0;tok[j];j++)
      {
        if (*tok[j])
        {
          strcat(vtok[i], tok[j]);
          strcat(vtok[i], " ");
        }
        free(tok[j]);
      }
      free(tok);
      vtok[i][strlen(vtok[i])-1] = 0;
      vtok[i] = (char *)realloc(vtok[i], strlen(vtok[i])+1);
    }
  }
  
  memset(b, 0, sizeof(b));
  for (i=0;nm[i]&&nm[i]!='(';i++)
    b[i] = nm[i];
  
  for (i=0;;i++)
  {
    if (!vars[i].nm)
    {
      if (vtok)
      {
        for (j=0;j<vcnt;j++)
          free(vtok[j]);
        free(vtok);
      }
      break;
    }
    if (!strcasecmp(vars[i].nm, b))
    {
      r = vars[i].func(vtok, vcnt);
      if (vtok)
      {
        for (j=0;j<vcnt;j++)
          free(vtok[j]);
        free(vtok);
      }
      break;
    }
  }
  
  if (!r)
    return(strdup(""));
  else  
    return(r);
}

char **form_vtoks(char *t, int *n)
{
  int i, j, k, c=0;
  char **ret;
  
  if (!strlen(t))
  {
    *n = 0;
    return(NULL);
  }
  
  ret = (char **)malloc(4096);
  
  for (i=0,j=0;;i++)
  {
    ret[i] = (char *)malloc(1024);
    memset(ret[i], 0, 1024);
    while (t[j] && t[j] == ' ')
      j++;
    for (k=0,c=0;t[j];k++,j++)
    {
      if (t[j] == ',' && !c)
        break;
      if (c < 0)
        c = 0;
      if (t[j] == '(')
        c++;
      else if (t[j] == ')')
        c--;
      ret[i][k] = t[j];
    }
    ret[i] = (char *)realloc(ret[i], strlen(ret[i])+1);
    if (!t[j])
      break;
    j++;
  }
  
  ret[i+1] = NULL;
  *n = i+1;
  
  return(ret);
}

unsigned char *doalias(alias_t *al, char **tok, int cnt)
{
  char *ret, echr;
  char **atok, p[16], pt[2048], tb[2048], *t;
  int i, j, acnt, n, l, c;
  
  t = strdup(al->args);
  
  for (i=0;al->args[i]&&al->args[i] == ' ';i++);
  
  if (al->args[i] == '{')
  {
    for (i=0,c=0;t[i];i++)
    {
      if (c < 0)
        c = 0;
      if (t[i] == '{' && !c)
      {
        t[i] = ' ';
        c++;
      }
      else if (t[i] == '{')
        c++;
      else if (t[i] == '}' && c == 1)
      {
        t[i] = ' ';
        c--;
      }
      else if (t[i] == '}')
        c--;
    }
  }
  
  if (t[strlen(t)-1] == '|' || t[strlen(t)-1] == '\n')
    t[strlen(t)-1] = 0;
  
  memset(pt, 0, sizeof(pt));
  strcpy(pt, t);
  free(t);
  
  for (i=0;pt[i];i++)
  {
    if (pt[i] == '$' || pt[i] == '?')
    {
      for (j=i,echr=0;;j++)
      {
        if (!pt[j] || pt[j] == ',' || pt[j] == ' ' || pt[j] == '=' || pt[j] == '!' || pt[j] == '>' || pt[j] == '<')
        {
          echr = pt[j];
          break;
        }
      }
      if (i > 0 && (pt[i-1] == '(' || pt[i-1] == ')' || pt[i-1] == ',' || pt[i-1] == '=' || pt[i-1] == '!' || pt[i-1] == '<' || pt[i-1] == '>' || pt[i-1] == ' '))
      {
        if (pt[i-1] != ' ')
        {
          memset(tb, 0, sizeof(tb));
          strcpy(tb, pt+i);
          strcpy(pt+i+1, tb);
          pt[i] = ' ';
        }
        else
          j--;
        if (echr && echr != ' ')
        {
          memset(tb, 0, sizeof(tb));
          strcpy(tb, pt+j+1);
          strcpy(pt+j+2, tb);
          pt[j+1] = ' ';
        }
      }
    }
  }
  
  t = strdup(pt);
  
  atok = form_tokso(t, &acnt);
  free(t);
  
  for (i=0,j=0;i<acnt;i++)
    if (*atok[i] == '$' && atok[i][1])
      if (atoi(atok[i]+1) > j)
        j = atoi(atok[i]+1);

  if ((cnt-1) < j)
    return(NULL);
  
  for (i=0;atok[i];i++)
  {
    if (*atok[i] == '?' && isdigit(atok[i][1]))
    {
      if (atoi(atok[i]+1) <= (cnt-1))
        *atok[i] = '$';
      else
      {
        for (j=1;;j++)
        {
          if (!isdigit(atok[i][j]) && atok[i][j] != '-')
          {
            memset(tb, 0, sizeof(tb));
            strcpy(tb, atok[i]+j);
            strcpy(atok[i], tb);
            break;
          }
        }
      }
    }
  }
  
  for (i=0,l=0;atok[i];i++)
  {
    if (*atok[i] == '$' && isdigit(atok[i][1]) && !strchr(atok[i], '-'))
    {
      n = atoi(atok[i]+1);
      memset(tb, 0, sizeof(tb));
      for (j=1;atok[i][j];j++)
        if (!isdigit(atok[i][j]))
          break;
      strcpy(tb, atok[i]+j);
      atok[i] = (char *)realloc(atok[i], strlen(tok[n])+strlen(tb)+1);
      strcpy(atok[i], tok[n]);
      strcat(atok[i], tb);
    }
    else if (*atok[i] == '$' && isdigit(atok[i][1]))
    {
      memset(p, 0, sizeof(p));
      for (j=1;;j++)
      {
        if (!isdigit(atok[i][j]) && atok[i][j] != '-')
        {
          memset(tb, 0, sizeof(tb));
          strcpy(tb, atok[i]+j);
          strcpy(atok[i], tb);
          break;
        }
        if (atok[i][j] != '-')
          p[j-1] = atok[i][j];
      }
      n = atoi(p);
      memset(pt, 0, sizeof(pt));
      for (j=n;tok[j];j++)
      {
        strcat(pt, tok[j]);
        strcat(pt, " ");
      }
      pt[strlen(pt)-1] = 0;
      atok[i] = (char *)realloc(atok[i], strlen(pt)+strlen(tb)+1);
      strcpy(atok[i], pt);
      strcat(atok[i], tb);
    }
    else if (*atok[i] == '$' && !strcasecmp(atok[i]+1, "num"))
    {
      free(atok[i]);
      atok[i] = NULL;
      msprintf(&atok[i], "%i", cnt-1);
    }
    else if (*atok[i] == '$' && !strcasecmp(atok[i]+1, "str"))
    {
      free(atok[i]);
      atok[i] = (char *)malloc(4096);
      memset(atok[i], 0, 4096);
      for (j=0;tok[j];j++)
      {
        if (*tok[j])
        {
          strcat(atok[i], tok[j]);
          strcat(atok[i], " ");
        }
      }
      atok[i][strlen(atok[i])-1] = 0;
      atok[i] = (char *)realloc(atok[i], strlen(atok[i])+1);
    }
    l+=strlen(atok[i])+1;
  }

  ret = (char *)malloc(4096);
  memset(ret, 0, 4096);
  
  for (i=0;atok[i];i++)
  {
    if (*atok[i])
    {
      strcat(ret, atok[i]);
      strcat(ret, " ");
    }
    free(atok[i]);
  }
  
  free(atok);
  
  ret[strlen(ret)-1] = 0;
  ret = (char *)realloc(ret, strlen(ret)+1);
  
  return((unsigned char *)ret);
}

int savealiases(char *fn)
{
  FILE *f;
  alias_t *cur;
  
  f = fopen(fn, "w");
  if (!f)
    return(-1);
  
  for (cur=alhead;cur;cur=cur->next)
    fprintf(f, "%s %s\n", cur->nm, cur->args);
  
  fclose(f);
  
  return(1);
}

int loadaliases(char *fn)
{
  FILE *f;
  alias_t *cur, *cur1=NULL;
  char buf[4096], tb[4096];
  int i, c, j;
  
  f = fopen(fn, "r");
  if (!f)
    return(-1);
  
  for (cur=alhead;cur;cur=cur->next)
    cur1 = cur;
  
  while (!feof(f))
  {
    memset(buf, 0, sizeof(buf));
    fgets(buf, sizeof(buf), f);
    if (*buf == '#')
      continue;
    if (buf[strlen(buf)-1] == '\n')
      buf[strlen(buf)-1] = 0;
    for (i=0,c=0;buf[i];i++)
    {
      if (buf[i] == ' ')
      {
        c = 1;
        break;
      }
    }
    
    if (!c)
    {
      buf[strlen(buf)] = ' ';
      fgets(buf+strlen(buf), sizeof(buf), f);
      if (buf[strlen(buf)-1] == '\n')
        buf[strlen(buf)-1] = 0;
    }
    
    buf[strlen(buf)] = '\n';
    
    for (c=0,i=0;;i++)
    {
      if (c < 0)
      {
        c = 0;
        break;
      }
      if (!buf[i] && c)
      {
        while (1)
        {
          memset(tb, 0, sizeof(tb));
          fgets(tb, sizeof(tb), f);
          for (j=0;tb[j]==' ';j++);
          if (tb[j] != '#' || feof(f))
          {
            strcpy(buf+i+1, tb);
            buf[i] = ' ';
            break;
          }
        }
      }
      else if (!buf[i])
        break;
      if (buf[i] == '{')
        c++;
      else if (buf[i] == '}')
        c--;
    }
    
    for (i=0;buf[i];i++)
    {
      if (buf[i] == '\n')
      {
        for (j=i+1;buf[j]==' '&&buf[j]!='{';j++);
        if (buf[j] == '{')
        {
          buf[i] = ' ';
          strncpy(buf+i+1, buf+j, strlen(buf+j));
          buf[strlen(buf)-(j-i)+1] = 0;
        }
      }
    }
    if (*buf == ' ')
      continue;
    for (i=strlen(buf)-1;buf[i] == ' '&&i>=0;i--);
    while (buf[i] == '\n')
    {
      buf[i] = 0;
      i--;
    }
    cur = (alias_t *)malloc(sizeof(alias_t));
    for (i=0;buf[i]!=' ';i++);
    cur->args = strdup(buf+i+1);
    cur->nm = (char *)malloc(i);
    cur->next = NULL;
    if (!cur1)
      alhead = cur;
    else
      cur1->next = cur;
    cur1 = cur;
    strncpy(cur->nm, buf, i);
    cur->nm[i] = 0;
  }
  
  fclose(f);
  
  return(1);
}

int loadsets(char *fn)
{
  FILE *f;
  char buf[4096], *nm;
  int i, j;
  unsigned char t;
  
  f = fopen(fn, "r");
  if (!f)
    return(-1);
  
  while (!feof(f))
  {
    memset(buf, 0, sizeof(buf));
    fgets(buf, sizeof(buf), f);
    if (!strchr(buf, ' '))
      continue;
    if (buf[strlen(buf)-1] == '\n')
      buf[strlen(buf)-1] = 0;
    for (i=0;buf[i]!=' ';i++);
    nm = (unsigned char *)malloc(i+1);
    strncpy(nm, buf, i);
    nm[i] = 0;
    i++;
    for (t=0,j=i;buf[i];i++)
    {
      if (!isdigit(buf[i]))
      {
        t = 1;
        break;
      }
    }
    addset(nm, buf+j, t);
    free(nm);
  }
  
  fclose(f);
  
  return(1);
}

int savesets(char *fn)
{
  FILE *f;
  sets_t *cur;
  
  f = fopen(fn, "w");
  if (!f)
    return(-1);
  
  for (cur=setl;cur;cur=cur->next)
    fprintf(f, "%s %s\n", cur->nm, cur->d);
  
  fclose(f);
  
  return(1);
}

void addset(unsigned char *nm, unsigned char *val, unsigned char t)
{
  sets_t *cur, *cur1=NULL;
  
  if (!setl)
  {
    setl = (sets_t *)malloc(sizeof(sets_t));
    cur = setl;
  }
  else
  {
    for (cur=setl;cur;cur=cur->next)
      cur1 = cur;
    cur = (sets_t *)malloc(sizeof(sets_t));
    cur1->next = cur;
  }
  
  cur->nm = strdup(nm);
  cur->d = strdup(val);
  cur->t = t;
  cur->next = NULL;
}

void chset(unsigned char *nm, unsigned char *d)
{
  unsigned char t=0;
  int i;
  
  delset(nm);
  
  for (i=0;d[i];i++)
  {
    if (isalpha(d[i]))
    {
      t = 1;
      break;
    }
  }
  
  if (!strlen(d))
    t = 1;
  
  addset(nm, d, t);
}

sets_t *findset(unsigned char *nm)
{
  sets_t *cur;
  
  for (cur=setl;cur&&strcasecmp(cur->nm, nm);cur=cur->next);
  
  return(cur);
}

unsigned char *getval(unsigned char *nm, unsigned char *t)
{
  sets_t *cur;
  
  cur = findset(nm);
  if (!cur)
  {
    *t = 0;
    return(NULL);
  }
  
  *t = cur->t;
  
  return(cur->d);
}

void delset(unsigned char *nm)
{
  sets_t *cur, *cur1=NULL;
  
  for (cur=setl;cur&&strcasecmp(cur->nm, nm);cur=cur->next)
    cur1 = cur;
  
  if (!cur)
    return;
  
  if (cur1)
    cur1->next = cur->next;
  else if (cur->next)
    setl = cur->next;
  else
    setl = NULL;
  
  free(cur->d);
  free(cur->nm);
  free(cur);
}

long nvar(unsigned char *nm)
{
  sets_t *cur;
  
  cur = findset(nm);
  if (!cur)
    return(-1);
  
  if (cur->t)
    return(-1);
  
  return(atol(cur->d));
}

char *strev(char *str, unsigned char l, unsigned char r, char *(*func)())
{
  char *s1, *s2, *rt, *t, *t2;
  char t3[1024], t4[1024], **tok, echr;
  int i, j, cnt;
  
  memset(t3, 0, sizeof(t3));
  strcpy(t3, str);
  
  for (i=0;t3[i];i++)
  {
    if (t3[i] == '$')
    {
      for (j=i,echr=0;;j++)
      {
        if (!t3[j] || t3[j] == ',' || t3[j] == ' ' || t3[j] == '=' || t3[j] == '!' || t3[j] == '>' || t3[j] == '<')
        {
          echr = t3[j];
          break;
        }
      }
      if (i > 0 && (t3[i-1] == '(' || t3[i-1] == ')' || t3[i-1] == ',' || t3[i-1] == '=' || t3[i-1] == '!' || t3[i-1] == '<' || t3[i-1] == '>' || t3[i-1] == ' '))
      {
        if (t3[i-1] != ' ')
        {
          memset(t4, 0, sizeof(t4));
          strcpy(t4, t3+i);
          strcpy(t3+i+1, t4);
          t3[i] = ' ';
        }
        else
          j--;
        if (echr && echr != ' ')
        {
          memset(t4, 0, sizeof(t4));
          strcpy(t4, t3+j+1);
          strcpy(t3+j+2, t4);
          t3[j+1] = ' ';
        }
      }
    }
  }
  
  if (strlen(t3) > 3 && t3[strlen(t3)-1] == ')' && t3[strlen(t3)-2] == ')')
  {
    t3[strlen(t3)-1] = ' ';
    t3[strlen(t3)] = ')';
  }
  
  t3[strlen(t3)-1] = ' ';
  t3[strlen(t3)] = ')';
  
  tok = form_tokso(t3, &cnt);
  tok = fxv(tok, cnt, &cnt);
  
  for (i=0;tok[i];i++)
  {
    if (*tok[i] == '$' && tok[i][1] && !isdigit(tok[i][1]))
    {
      t = dovars(tok[i]+1);
      if (t)
      {
        free(tok[i]);
        tok[i] = t;
      }
    }
  }
  
  memset(t3, 0, sizeof(t3));
  
  for (i=0;tok[i];i++)
  {
    if (*tok[i])
    {
      strcat(t3, tok[i]);
      strcat(t3, " ");
    }
    free(tok[i]);
  }
  
  free(tok);
  
  if (t3[strlen(t3)-1] == ' ')
    t3[strlen(t3)-1] = 0;
  
  if (t3[strlen(t3)-1] != r)
  {
    t3[strlen(t3)] = ' ';
    t3[strlen(t3)] = r;
    t3[strlen(t3)] = 0;
  }
  
  rt = strdup(t3);
  
  for (;;)
  {
    s1 = strrchr(rt, l);
    if (s1 == NULL)
      break;
    if (strchr(rt, r) == NULL)
    {
      free(rt);
      return(NULL);
    }
    s2 = (char *)malloc(abs(strchr(s1, r)-s1));
    strncpy(s2, s1+1, abs(strchr(s1, r)-s1)-1);
    s2[abs(strchr(s1, r)-s1)-1] = '\0';
    t = func(s2);
    if (t == NULL)
    {
      free(rt);
      free(s2);
      return(NULL);
    }
    t2 = strdup(rt);
    free(rt);
    s1 = strrchr(t2, l);
    rt = ins(t2, t, abs(s1-t2), abs((s1-t2))+(abs((strchr(s1, r)-s1))+1));
    free(t2);
    free(t);
  }
  
  return(rt);
}

char *cmp(char *st)
{
  char b[512], r[512], **tok, *tm=NULL, *rt=NULL;
  int i=0, j, cnt, k, l=0, t;
  
  memset(b, 0, sizeof(b));
  memset(r, 0, sizeof(r));
  
  while (st[i])
  {
    while (st[i] == ' ' && st[i])
      i++;
    for (b[0]='\'',j=1;st[i]&&st[i]!='&'&&st[i]!='^';j++,i++)
    {
      if (st[i] == '=' || st[i] == '!' || st[i] == '>' || st[i] == '<')
      {
        b[j++] = '\'';
        b[j++] = ' ';
        b[j++] = st[i];
        b[j++] = ' ';
        b[j] = '\'';
      }
      else
        b[j] = st[i];
    }
    b[j] = '\'';
    b[j+1] = 0;
    if (st[i] == '&' || st[i] == '^')
      i++;
    tok = form_toks(b, &cnt);
    tok = fxv(tok, cnt, &cnt);
    
    if (!cnt)
    {
      tok[0] = NULL;
      tok[1] = NULL;
      tok[2] = NULL;
    }
    if (cnt == 1)
    {
      tok[1] = NULL;
      tok[2] = NULL;
    }
    else if (cnt == 2)
      tok[2] = NULL;
    
    if (tok[0] && *tok[0] == '$')
    {
      tm = dovars(tok[0]+1);
      if (tm)
      {
        free(tok[0]);
        tok[0] = tm;
      }
    }
    if (tok[1] && *tok[1] == '$')
    {
      tm = dovars(tok[1]+1);
      if (tm)
      {
        free(tok[1]);
        tok[1] = tm;
      }
    }
    if (tok[2] && *tok[2] == '$')
    {
      tm = dovars(tok[2]+1);
      if (tm)
      {
        free(tok[2]);
        tok[2] = tm;
      }
    }
    
    tm = NULL;
    if (!tok[0])
      tm = strdup("");
    else
      msprintf(&tm, "\'%s\' %s \'%s\'", tok[0]?tok[0]:"", tok[1]?tok[1]:"", tok[2]?tok[2]:"");
    for (k=0;k<cnt;k++)
      free(tok[k]);
    free(tok);
    tok = form_toks(tm, &cnt);
    free(tm);
    
    if (!cnt)
    {
      tok[0] = NULL;
      tok[1] = NULL;
      tok[2] = NULL;
    }
    if (cnt == 1)
    {
      tok[1] = NULL;
      tok[2] = NULL;
    }
    else if (cnt == 2)
      tok[2] = NULL;
    
    if (tok[2] && !(*tok[2]))
    {
      free(tok[2]);
      tok[2] = NULL;
      cnt--;
    }
    if (tok[1] && !(*tok[1]))
    {
      free(tok[1]);
      tok[1] = NULL;
      if (tok[2])
      {
        tok[1] = strdup(tok[2]);
        free(tok[2]);
        tok[2] = NULL;
      }
      cnt--;
    }
    if (tok[0] && !(*tok[0]))
    {
      free(tok[0]);
      if (tok[1])
      {
        tok[0] = strdup(tok[1]);
        free(tok[1]);
        tok[1] = NULL;
      }
      if (tok[2])
      {
        tok[1] = strdup(tok[2]);
        free(tok[2]);
        tok[2] = NULL;
      }
      cnt--;
    }
    
    if (cnt == 2)
    {
      if (*tok[0] == '=' || *tok[0] == '!' || *tok[0] == '>' || *tok[0] == '<')
      {
        if (*tok[0] == '!')
        {
          if (strcmp(tok[1], "0"))
          {
            free(tok[0]);
            free(tok[1]);
            tok[3] = NULL;
            tok[2] = strdup("0");
            tok[1] = strdup("!");
            tok[0] = strdup("0");
          }
          else
          {
            free(tok[0]);
            free(tok[1]);
            tok[3] = NULL;
            tok[2] = strdup("0");
            tok[1] = strdup("=");
            tok[0] = strdup("0");
          }
        }
        else
        {
          tok[3] = NULL;
          tok[2] = strdup(tok[1]);
          free(tok[1]);
          tok[1] = strdup(tok[0]);
          free(tok[0]);
          tok[0] = strdup("0");
        }
      }
      else if (*tok[1] == '=' || *tok[1] == '!' || *tok[1] == '>' || *tok[1] == '<')
      {
        tok[3] = NULL;
        tok[2] = strdup(tok[0]);
        if (*tok[1] == '=')
          *tok[1] = '!';
        else if (*tok[1] == '!')
          *tok[1] = '=';
      }
      cnt = 3;
    }
    else if (cnt == 1)
    {
      if (strcmp(tok[0], "0") || *tok[0] == '!')
      {
        free(tok[0]);
        tok[0] = strdup("1");
        tok[1] = strdup("=");
        tok[2] = strdup("1");
        tok[3] = NULL;
      }
      else
      {
        free(tok[0]);
        tok[0] = strdup("1");
        tok[1] = strdup("!");
        tok[2] = strdup("1");
        tok[3] = NULL;
      }
      cnt = 3;
    }
    else if (!cnt)
    {
      tok[0] = strdup("1");
      tok[1] = strdup("!");
      tok[2] = strdup("1");
      tok[3] = NULL;
      cnt = 3;
    }
    
    if (cnt == 3)
    {
      if (*tok[1] == '=')
      {
        if (!strcasecmp(tok[0], tok[2]))
          r[l++] = 1;
        else
          r[l++] = 0;
      }
      else if (*tok[1] == '!')
      {
        if (!strcasecmp(tok[0], tok[2]))
          r[l++] = 0;
        else
          r[l++] = 1;
      }
      else if (*tok[1] == '>')
      {
        if (atoi(tok[0]) > atoi(tok[2]))
          r[l++] = 1;
        else
          r[l++] = 0;
      }
      else if (*tok[1] == '<')
      {
        if (atoi(tok[0]) < atoi(tok[2]))
          r[l++] = 1;
        else
          r[l++] = 0;
      }
    }
    
    for (k=0;tok[k];k++)
      free(tok[k]);
    free(tok);
    
    if (i > 0 && (st[i-1] == '&' || st[i-1] == '^'))
      r[l++] = st[i-1];
  }
  
  r[l] = 0;
  
  for (i=0,t=0;i<l;i++)
  {
    if (r[i] == '&')
      t &= r[++i];
    else if (r[i] == '^')
      t |= r[++i];
    else
      t = r[i];
  }
  
  msprintf(&rt, "%i", t);
  
  return(rt);
}
